/*------------------------------------------------------------------------
Module:        mime_alter.c
Author:        Paul L Daniels / Steven Coutts
Project:       ripMIME
State:         development
Creation Date: 01/05/2001
Description:   MIME-alter is an additional module in the mime.[ch] object, which will contain various functions which will -alter- mime-packs.
------------------------------------------------------------------------*/

// 0908 10H57 - Added ability for disclaimer to be actual text, as well as
// 	a filename.  I just detect for a \n or \r in the string, if detected
//		disclaimer is considered as "text"
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <syslog.h>
#include <errno.h>


/* Our content types */
#define _unknown -1
#define _none 0
#define _text_plain 1
#define _text_html 2
#define _text_print 3
#define _text_unknown 4

/* Our encoding scemes */
#define _base64 1
#define _print 2
#define _7bit 3

#define _BUFFER_MAX 1024

/* Globals */
char *buffer[_BUFFER_MAX]; /* allocate a static array of 1024 pointers to char */
unsigned char encode64[64]; /* binary -> base64 encoding table */

int __verbose=0;				/* do we talk as we walk */
int __paranoid=1;				/* set paranoid to yes! */
int __HTMLtoo=1;				/* Add footer to the HTML email too */
int __nullify_all=0;			/* Remove ALL filename'd attachments */
int __alter_signed=0;		/* Do we alter signed emails ? */

int _AM_syslogging=1;		/* do we want to log to syslog */
int _AM_stderrlogging=0;	/* do we want to log to stderr */

struct _theaders
{
	int content_type;
	int content_encoding;
	char filename[1024];
	char boundary[1024];
};

#define _HEADERBUFFER_MAX 100

char *headerbuffer[_HEADERBUFFER_MAX];	// 100 lines for the header buffers

int headerbuffermax=0;
int _debugAM = 0;




/*------------------------------------------------------------------------
 Procedure:     AM_set_debug ID:1
 Purpose:
 Input:
 Output:
 Errors:
------------------------------------------------------------------------*/
int AM_set_debug( int level )
{
	_debugAM = level;

	return 1;
}


/*------------------------------------------------------------------------
Procedure:     AM_hbuffer_reset ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_hbuffer_reset( void )
{
	int i;

	for (i = 0; i < headerbuffermax; i++)
	{
		if (headerbuffer[i]) free(headerbuffer[i]);
	}
	headerbuffermax=0;
	return 1;
}


/*------------------------------------------------------------------------
Procedure:     AM_hbuffer_add ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
char *AM_hbuffer_add( char *headerline, FILE *f )
{
	int i;

	if (headerbuffermax == _HEADERBUFFER_MAX)
	{
		if (headerbuffer[0])
		{
			fprintf(f,"%s",headerbuffer[0]);
			free(headerbuffer[0]);
		}
		for (i = 0; i < (_HEADERBUFFER_MAX-1); i++)
		{
			headerbuffer[i] = headerbuffer[i+1];
		}
		headerbuffermax = _HEADERBUFFER_MAX -1;
	}

	headerbuffer[headerbuffermax] = malloc(sizeof(char) *(1025));

	if (!headerbuffer[headerbuffermax])
	{
		syslog(1,"alterMIME: AM_hbuffer_add: Error: cannot allocate %d bytes for new header",sizeof(headerline)+1);
		return NULL;
	}

	strncpy(headerbuffer[headerbuffermax], headerline, 1024);
	headerbuffermax++;

	return headerbuffer[headerbuffermax-1];
}


/*------------------------------------------------------------------------
Procedure:     AM_hbuffer_getmax ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_hbuffer_getmax( void )
{
	return headerbuffermax;
}




/*------------------------------------------------------------------------
Procedure:     AM_hbuffer_getline ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
char *AM_hbuffer_getline( int index )
{
	if ((index >=0)&&(index < headerbuffermax))
		return headerbuffer[index];
	else return NULL;
}



/*------------------------------------------------------------------------
Procedure:     AM_set_paranoid ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_set_paranoid( int level )
{
	__paranoid = level;
	return __paranoid;
}





/*------------------------------------------------------------------------
Procedure:     AM_set_verbose ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_set_verbose( int level )
{
	__verbose = level;
	return __verbose;
}




/*------------------------------------------------------------------------
Procedure:     AM_set_HTMLtoo ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_set_HTMLtoo( int level )
{
	__HTMLtoo = level;
	return __HTMLtoo;
}




/*------------------------------------------------------------------------
Procedure:     AM_set_nullifyall ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_set_nullifyall( int level )
{
	__nullify_all = level;
	return __nullify_all;
}



/*------------------------------------------------------------------------
 Procedure:     AM_set_altersigned ID:1
 Purpose:       Turns on or off the option to alter signed email messages.
                Default is to be -off-.
 Input:
 Output:
 Errors:
------------------------------------------------------------------------*/
int AM_set_altersigned( int level )
{
	__alter_signed = level;
	return __alter_signed;
}




/*------------------------------------------------------------------------
Procedure:     strlower ID:1
Purpose:       Converts a given string into all lowercase
Input:         convertme : String to be converted
Output:        none
Errors:
------------------------------------------------------------------------*/
int AM_strlower( char *convertme )
{

	char *c = convertme;

	while ( *c != '\0') {*c = tolower(*c); c++;}

	return 0;
	}




/*------------------------------------------------------------------------
Procedure:     AM_create_encode_table ID:1
Purpose:       Builds the value:ASCII relationship table for BASE64 -encoding-
process
Input:         none
Output:
Errors:
------------------------------------------------------------------------*/
int AM_create_encode_table( void )
{

	int i,j;

	for ( i = 'A', j=0; i <= 'Z'; i++, j++)
	{
		encode64[j]=i;
	}

	for ( i = 'a'; i <= 'z'; i++, j++)
	{
		encode64[j]=i;
	}

	for ( i = '0'; i <= '9'; i++, j++)
	{
		encode64[j]=i;
	}

	encode64[62] = '+';
	encode64[63] = '/';
	return 0;
}







/*------------------------------------------------------------------------
Procedure:     AM_base64_encodef ID:1
Purpose:       encode to BASE64 an input stream to an output stream
Input:         FILE *fin: Input stream
FILE *fout: Output stream
Output:
Errors:
------------------------------------------------------------------------*/
int AM_base64_encodef( FILE *fin, FILE *fout )
{
	unsigned char inbuf[3];
	unsigned char outbuf[4];
	int cc;
	int byte_count;

	if (!fin)
	{
		syslog(1,"AM_base64_encodef: Error: input file stream not open, please use AM_base64_encode(infile,outfile)");
		return -1;
	}

	if (!fout)
	{
		syslog(1,"AM_base64_encodef: Error: output file stream not open, please use AM_base64_encode(infile,outfile)");
		return -1;
	}

	AM_create_encode_table(); /* setup the binary->base64 encoding table */

	cc = 0;	/* Character Count = 0 */

	while ((byte_count = fread(inbuf, 1, 3, fin)))	/* while we get more than 0 bytes */
	{
		/* Split 3 bytes into 4 bytes by taking 6bit blocks out of the 24 bit (3x8bit)
		* array */
		outbuf[0] = encode64[((inbuf[0] & 0xFC) >> 2)];
		outbuf[1] = encode64[((inbuf[0] & 0x03) << 4) | ((inbuf[1] & 0xF0) >> 4)];
		outbuf[2] = encode64[((inbuf[1] & 0x0F) << 2) | ((inbuf[2] & 0xC0) >> 6)];
		outbuf[3] = encode64[(inbuf[2] & 0x3F)];

		/* if we didn't get a full 3 bytes from the file read, then we need to then
		* pad the output with '=' (base64 padding char) */
		if ( byte_count < 3 )
		{

			if ( byte_count == 2 ) outbuf[3] = '=';
			if ( byte_count == 1 ) outbuf[3] = outbuf[2] = '=';

			fwrite(outbuf, 1, 4, fout); 	/* write out the buffer */
			fwrite( "\n", 1, 1, fout ); 	/* as this is the last line, put a trailing \n */
			break; 								/* exit out of the while loop */
		}

		fwrite(outbuf, 1, 4, fout);		/* normal operation buffer-write */

		cc += 4;									/* increment the output char count by 4 */

		if (cc > 76)							/* if we have more than 76 chars, then put in a \n */
		{
			fwrite( "\n", 1, 1, fout );	/* write the \n out */
			cc = 0;								/* reset the output char count */
		} /* if bytecount == 3 */
	} /* while more data to read in */

	return 0;
}






/*------------------------------------------------------------------------
Procedure:     AM_base64_encode ID:1
Purpose:       Encodes a given input stream into BASE64 and outputs to a given
output stream
Input:         char *enc_fname: Input filename of the stream to encode
char *out_fname: Output filename of the stream to send to
Output:
Errors:
------------------------------------------------------------------------*/
int AM_base64_encode( char *enc_fname, char *out_fname )
{
	FILE *fin, *fout;

	fin = fopen( enc_fname, "rb" );
	if (!fin)
	{
		syslog(1,"AM_base64_encode: Cannot open \"%s\" for reading.(%s)",enc_fname,strerror(errno));
		return -1;
	}

	fout = fopen ( out_fname, "wb" );
	if (!fout)
	{
		syslog(1,"AM_base64_encode: Cannot open \"%s\" for writing.(%s)",out_fname,strerror(errno));
		fclose(fin);
		return -1;
	}

	AM_base64_encodef( fin, fout );	/* encode and output */

	fclose(fin);	/* close the input file */
	fclose(fout);	/* close the output file */

	return 0;
}




/*------------------------------------------------------------------------
Procedure:     AM_quick_clean_filename ID:1
Purpose:       scans all the chars of a filename, converts any non-safe chars to
a '_'
Input:         char *filename: The filename we're going to convert/scan
Output:
Errors:
------------------------------------------------------------------------*/
void AM_quick_clean_filename( char *fname )
{
	char tmp[1024];
	char *p = strrchr(fname,'/');

	/* scan out any directory separators */
	if (p)
	{
		p++;
		sprintf(tmp,"%s",p);
		strcpy(fname,p);
	}
	else if ( (p = strrchr(fname,'\\')))
	{
		p++;
		sprintf(tmp,"%s",p);
		strcpy(fname,p);
	}


	while (*fname) {
		if ( !isalnum(*fname)||(*fname == ' ') ) *fname='_';
		fname++;
	}
}






/*------------------------------------------------------------------------
Procedure:     AM_attachment_replace ID:1
Purpose:       Replace the attachment in a mailbody
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_attachment_replace( char *mpack_name, char *old_att_name, char *new_att_name, char *att_path )
{
	FILE *in, *tmp, *newa;
	char *ofn, *nfn;
	char line[1024];
	char lline[1024];
	char lname[1024];
	char ts[1024];
	char tname[1024];
	char *p, *q;
	int base64hit = 0;
	int hit = 0;
	int quote_on = 0;
	int namesmatch = 0;
	int i = 0;
	char *index;
	char *nameindex = NULL;
	char *fnameindex = NULL;

	in = fopen( mpack_name, "r" ); /* open up our original mailpack */

	if (!in)	/* handle file opening error */
	{
		syslog(1,"AM_attachment_replace: Cannot open mailpack %s for reading, \"%s\""\
		,mpack_name\
		,strerror(errno));
		return -1;
	}


	// First we must take the old file name, and translate it into
	// something we can use to compare against after it has been
	// brutalised and reduced (for safety)
	//
	ofn = strrchr(old_att_name,'/');  	/* get the attachment name out of the full path */
	if (!ofn) ofn = old_att_name;			/* if there was no '/', then reset the ofn ptr */
	else ofn++; 								/* skip the last '/' */
	strcpy(lname, ofn); 						/* create a lowercase version */
	AM_strlower(lname);
	AM_quick_clean_filename(lname); 		/* clean up the filename */

	// Get the new attachment name out of a full path
	// This is simply so that we can insert this new name
	// into the headers without the pointless path prefix.
	//
	nfn = strrchr(new_att_name,'/');		/* extract new att name from path */
	if (!nfn) nfn = new_att_name;			/* if there was no '/' then reset the nfn ptr */
	else nfn++;


	if (_debugAM) syslog(1,"AM_replace(): looking to replace %s with %s, having %s and %s", ofn, nfn, mpack_name, att_path );

	// Create the temporary mailpack name
	// This will become the new mail pack
	// assuming that everything goes okay in the end.
	//
	sprintf(tname,"%s.tmp", mpack_name);/* create the temporary mailpack name */
	tmp = fopen( tname, "w" );				/* try open the temp file */
	if (!tmp) /* handle file opening error */
	{
		syslog(1,"AM_attachment_replace: Cannot open temporary file %s for writing, \"%s\""\
		,tname\
		,strerror(errno));
		return -1;
	}


	// Into the mailpack, first we must read in the headers, this must
	// be done in memory, so that should we encounter any of the required
	// header details, we dont have to try and back-track through a file
	// This gets rather awkward, as we dont know (often) if the file is
	// replacable before we get to its first occurance of the filename
	// header. This means we often have to backtrack ( a real PAIN ).
	//
	while (!feof(in))
	{
		hit = 0;
		base64hit = 0;
		quote_on = 0;
		namesmatch = 0;
		nameindex = NULL;
		fnameindex = NULL;
		AM_hbuffer_reset();// prepare the header buffers


		// Read in our file lines, searching for our attachment
		// which we are supposed to be replacing.
		//
		while (fgets(line, 1024, in))
		{

			// Store this line into the header-buffer for later
			// use.
			//
			index = AM_hbuffer_add(line, tmp);

			strcpy(lline,line); /* copy and convert to lowercase the line we've read */
			AM_strlower(lline);

			// If we detect "base64" on this line, then set the flag
			//
			if (strstr(lline,"base64"))
			{
				if (_debugAM) syslog(1,"AM_replace(): BASE64 found - flag set.");
				base64hit=1;
			}

			// If we detect "filename=" on this line, then we need to
			// remember it, for later processing (if required)
			//
			if ((p=strstr(lline,"name="))) /* if the line contains name=, then check it */
			{
				if (!fnameindex) fnameindex = index;
				else nameindex = index;

				// Okay, so we've found the "name" line, now, lets
				// process the data on this line, and see if we have a
				// matching filename.
				//
				quote_on = 0;					/* no quotes found yet */
				p = p +strlen("name=");		/* move to just after the ='s */
				if (*p == '\"')				/* if there's a quote... */
				{
					quote_on++;					/* set quote to ON */
					p++;							/* move to next char */
				}

				q = strtok(p,";\"\n\r");	/* extract the filename */
				AM_quick_clean_filename(q);/* clean the filename */

				// Compare both the length and contents of the
				// filenames.
				//
				if (\
					(strlen(q) == strlen(lname))\
					&& (strncmp(q,lname, strlen(lname))==0)\
					)
				{
					if (_debugAM) syslog(1,"AM_replace(): Hit on filenames  %s and %s",q,lname);
					namesmatch = 1;
				}
				else
				    {
					namesmatch = 0;
					base64hit = 0;
					fnameindex = 0;
					nameindex = 0;
				}
			}


			// Now, if we've found this is a base64 attachment and we
				// have also matched the names (matching done here)
				//
			if ( (base64hit == 1) && (namesmatch == 1) )
			{
				if (_debugAM) syslog(1,"Base64 and NameMatch located, now fixing headers");
				// Okay, so we've gotten the match, and we need to replace
					// this filename details with those of the new file...
					// This is were we use the nameindex and fnameindex, for
					// from those two points, we alter the filenames
					//

				// First, lets do the "name" only one
					// strcpy the details so we can play with it safely
					//
				if (nameindex)
				{
					strcpy(line, nameindex);
					AM_strlower(line);

					// 2. find where the "name" bit starts, and \0 it there
					//
					p = strstr(line,"name=") +strlen("name=");

					// 3. If the filename was quoted, remember this, and move one
					// 	char off to the right.
					if (*p == '"')
					{
						p++;
						quote_on=1;
					}
					else quote_on=0;
					*p = '\0';

					// 4. The actual filename is now ahead of us, until the next
					//		quote or \n\r.  We in effect, can ignore this now, and
					//		construct our new line.
					//
					sprintf(lline,"%s%s",line,nfn);
					if (quote_on) strcat(lline,"\"");
					strcat(lline,"\n");

					// 5. Copy this created string back, this requies a little
					//
					strncpy(nameindex,lline,1023);
				}

				//
					// The end of the filename swapout procedure.
					//
					// Now, do the same for the "filename" line....
					//
					if (fnameindex)
				{
					strcpy(line, fnameindex);
					AM_strlower(line);
					p = strstr(line,"name=") +strlen("name=");
					if (*p == '"')
					{
						p++;
						quote_on=1;
					}
					else quote_on=0;
					*p = '\0';
					sprintf(lline,"%s%s",line,nfn);
					if (quote_on) strcat(lline,"\"");
					strcat(lline,"\n");
					strncpy(fnameindex,lline,1023);
				}
				// We are now happy with the way the headers are in the
					// memory, we can dump them to the output file now.
					//
					hit++;						/* flag that we had success */
			}

			if (hit && base64hit) break;


		} // WHILE fgets(line)

		// Write out the headers which we have been so faithfully
		// keeping in our holds
		//
		for (i = 0;
		i < AM_hbuffer_getmax();
		i++)
		{
			fprintf(tmp,"%s",AM_hbuffer_getline(i));
		}

		// Reset the buffers
		//
		AM_hbuffer_reset();


		/* If we had success with finding the right filename etc... */
		if ((hit)&&(base64hit))
		{
			hit = base64hit = 0;

			/* Step 2. Locate the start of the attachment */
			/* Step 2. Locate the start of the attachment */
			/* Step 2. Locate the start of the attachment */
			while (fgets(line, 1024, in))
			{

				strcpy(lline,line);			/* copy and convert the next line */
				AM_strlower(lline);

				if ((p=strstr(lline,"name=")))	/* if the line has name= in it... */
				{

					p = p +strlen("name="); /* get the position where the name starts */
					if (*p == '\"')			/* if quoted... */
					{
						quote_on=1;				/* set quote flag */
						p++;						/* step one char to the right */
					}
					else quote_on=0;

					q = strtok(p,";\"\n\r");	/* extract the filename */
					AM_quick_clean_filename(q);/* clean the filename */

					if ( (strlen(q) == strlen(lname))\
				&& (strncmp(q,lname, strlen(lname))==0)
						    )
					{
						p = q;
						q = line +(p -lline);
						*q = '\0';
						sprintf(ts, "%s%s", line, nfn);
						if (quote_on) strcat(ts,"\"");
						strcat(ts,"\n");
						strcpy(line,ts);
					}
				}


				fprintf(tmp,"%s",line);

				if (strlen(line) < 3) break;	/* if we've gotten to a blank line, exit */
			}


			/* Step 3. Insert attachment */
			/* Step 3. Insert attachment */
			/* Step 3. Insert attachment */

			newa = fopen( att_path, "r" );		/* Open the new attachment */

			if (!newa)										/* if we couldn't open it */
			{
				syslog(1,"AM_attachment_replace: Cannot open file \"%s\" required to be inserted.(%s)"\
			,new_att_name, strerror(errno));

				/* as we couldn't open the new attachment, then we better keep the
				* existing file so that we dont bugger up the mailpack, better to have
				* the old attachment than non at all! */
				while (fgets(line, 1024, in))
				{
					fprintf(tmp,"%s",line);
					if (strpbrk(line,"-")) break;
				} // While getting input file

			} // if New Attachment couldn't be opened

			else
			    {
				if (_debugAM) syslog(1,"AM_replace(): Starting to encode file...");
				AM_base64_encodef( newa, tmp );	/* Encode and output the new attachment */
				fclose(newa);								/* close the new attachment */
				if (_debugAM) syslog(1,"AM_replace(): Encoding and insertion complete.");

				/* absorb the old attachment */
				while (fgets(line, 1024, in))
				{
					/* we're assuming that in base64, only
					* boundary lines have a '-' in them
					* This is possibly a bad thing (IMHO)
					* but short of tracking boundaries, I
					* have no other way */
					if (strpbrk(line,"-"))
					{
						fprintf(tmp,"%s",line);
						break;
					} // If strpbrk

				} // While fgets

				if (_debugAM) syslog(1,"AM_replace(): Attachment absorption complete.");

			} // ELSE

		} // While (major) fgets

	} // While !feof()

	clearerr(in);										/* clear out any errno's */
	fclose(tmp);										/* close our new mailpack */
	fclose(in);											/* close our old mailpack */
	rename( tname, mpack_name );					/* rename! */

	return 0;
}





/*------------------------------------------------------------------------
Procedure:     AM_insert_disclaimer ID:1
Purpose:       Inserts a disclaimer/footer file into a given email content stream,
on the basis that the content type/encoding are indicating that the
point in the email is a suitable (ie, non-binary-encoded) place.
Input:         inf - Input file stream, the disclaimer, as opened previously
outf - Output file stream, the new mailpack which is being created
content_type - indicates what content type was last detected
content_encoding - indicates what content encoding scheme was last detected
Output:        none
Errors:
------------------------------------------------------------------------*/
int AM_insert_disclaimer( FILE *inf, FILE *outf, int content_type, int content_encoding )
{
	char line[1024]="";

	if ((content_encoding == _none) || (content_encoding == _print) || (content_encoding == _7bit))
	{
		if (content_type == _text_html) fprintf(outf,"\n<PRE>\n");
		while (fgets(line,1024,inf)) fprintf(outf,"%s",line);
		fprintf(outf,"\n");
		if (content_type == _text_html) fprintf(outf,"</pre>\n");
	}

	return 0;
}




/*------------------------------------------------------------------------
Procedure:     AM_add_disclaimer ID:1
Purpose:       General purpose routine for adding what is typically a
disclaimer to email mailpacks.  This only inserts into the first
attachment body, or at the end of a non-encoded email
Input:         mpackname - The name of the mailpack which we are inserting into
disclaimer - The name of the file which contains the disclaimer
Output:        none
Errors:
------------------------------------------------------------------------*/
int AM_add_disclaimer( char *mpackname, char *disclaimer )
{
	FILE *mp, *d, *newf;
	char line[1024]="";
	char safe[1024]="";
	char lline[1024]="";
	char mpacktmp[1024]="";
	char mpackold[1024]="";
	char boundary[1024]="";

	int content_type = _none;
	int content_encoding = _none;
	int saveline = 0;
	int writeline = 0;
	int ishtml = 0;
	int isfile = 0;
	char *istext = strpbrk(disclaimer,"\n\r"); //Indicates if the dislcaimer is a filename, or the disclaimer itself!

	char *prebody, *tmpbody;

	int result = 0;
	int boundary_found = 0;
	int disclaimer_inserted = 0;
	int html_inserted = 0;
	char *pos, *newpos;

	/* create our temp filename */
	sprintf(mpacktmp,"%s.tmp",mpackname);
	sprintf(mpackold,"%s.old",mpackname);

	/* Initialise our files */
	newf = fopen(mpacktmp,"w");
	mp = fopen(mpackname,"r");


	// Test file statuses... and hop out if we had troubles with either
	//
	if (!newf)
	{
		syslog(1,"AM_add_disclaimer: Cannot open %s, %s",mpacktmp,strerror(errno));
		return -1;
	}

	if (!mp)
	{
		syslog(1,"AM_add_disclaimer: Cannot open %s, %s",mpacktmp,strerror(errno));
		return -1;
	}



	/* Initialise our variables */
	content_type = _none;
	content_encoding = _none;


	if (__verbose) fprintf(stderr,"Adding disclaimer...\n");


	/* Read every line from the original mailpack */
	while ( fgets(line,1024,mp) )
	{

		writeline = 0;

		/* copy and convert into lower-case the input line */
		strcpy(lline,line);
		strcpy(safe,line);
		AM_strlower(lline);

		/* check and see if this is a content-type line */
		if ((pos=strstr(lline, "content-type:")))
		{

			if (strstr(lline,"/signed"))
			{
				if (__alter_signed == 0)
				{
					if (__verbose) fprintf(stderr,"Message is signed, not altering.\n");
					fclose(mp);
					fclose(newf);
					return 0;
				}
			}
			writeline++;
			pos += strlen("content-type:");	/* move our pointer along */

			if ((pos=strstr(pos,"text/")))	/* do we have a text type */
			{
				pos += strlen("text/");	/* move pointer along */
				if (strncmp(pos, "plain",5)==0) content_type = _text_plain;
				else if (strncmp(pos, "html",4)==0) content_type = _text_html;
				else if (strncmp(pos, "print",5)==0) content_type = _text_print;
				else content_type = _text_unknown;
			}
			else
			    {
				content_type = _unknown;
			}

			if (__verbose) fprintf(stderr,"Content line found...\n");
		} /* if the line had content-type in it */


		/* Our transfer encoding system indicates if we have a attachment type which is
		* binary-encoded stuff or plain text */
		if ((pos=strstr(lline, "content-transfer-encoding:")))
		{
			writeline++;
			pos += strlen("content-transfer-encoding:");
			if (strstr(pos,"base64")) content_encoding = _base64;
			else if (strstr(pos,"quote")) content_encoding = _print;
			else if (strstr(pos,"7bit")) content_encoding = _7bit;
			if (__verbose) fprintf(stderr,"Content encoding found...\n");
		} /* if the line had content-transfer */



		pos = strstr(lline, "boundary=");	/* check for our boundary */
		if (pos)
		{
			if (__verbose) fprintf(stderr,"Possible boundary...");
			writeline++;
			newpos = (pos -lline) +line +strlen("boundary=");	/* move the pos pointer to just after the = */
			pos = newpos;
			if (*pos == '\"') pos++;	/* if our 1st char is a quote, move one more */
			newpos = strtok(pos,"\n\r\"");	/* tokenise out the boundary */
			strcpy(boundary, newpos);	/* copy it to our static string array */
			boundary_found++;	/* increment the boundary_found flag */
			if (__verbose) fprintf(stderr,"boundary located...");
		} /* if POS */


		if (writeline)
		{
			fprintf(newf,"%s",safe);
			continue;
		}


		// If we've found a boundary and a text content section...
		//
		if ((!disclaimer_inserted)\
		&&(boundary_found)\
		&&(content_type != _none)\
		&&(content_type != _unknown))
		{

			saveline = -1;

			while (fgets(line,1024,mp))
			{
				if (strstr(line,boundary)) break;
				fprintf(newf,"%s",line);
			}


			/* Add in the text only disclaimer */
			/* Add in the text only disclaimer */
			/* Add in the text only disclaimer */
			if (__verbose) fprintf(stdout,"Inserting text-body disclaimer...");

			if (istext)
			{
				fprintf(newf,"\n%s\n\n",disclaimer);
			}
			else
			    {
				d = fopen(disclaimer,"r");
				if (d)
				{
					AM_insert_disclaimer(d, newf, content_type, content_encoding);
					fclose(d);
				}
				else
				    {
					if (_AM_syslogging > 0) syslog(1,"AM_add_disclaimer: Error: Cannot open disclaimer file for reading (%s) : %s",disclaimer, strerror(errno));
					if (_AM_stderrlogging > 0) fprintf(stderr,"AM_add_disclaimer: Error: Cannot open disclaimer for reading");
				}
			}
			fprintf(newf,"%s",line);


			disclaimer_inserted = 1; /* Our disclaimer has been inserted, so set the flag */

			/* Add in the HTML disclaimer (if wanted) */
			/* Add in the HTML disclaimer (if wanted) */
			/* Add in the HTML disclaimer (if wanted) */
			if ((__HTMLtoo)&&(html_inserted == 0))
			{

				/* to ensure we dont try and insert this into a filenamed-attachment
				* we must scan each line for a "filename" until we find our </body>
				* tag */
				while (fgets(line,1024,mp))
				{
					strcpy(lline,line);
					AM_strlower(lline);

					if (strstr(lline,"content-type: text/html"))
					{
						ishtml = 1;
					}
					else if (strstr(lline,"filename="))
					{
						isfile = 1;
					}

					if ((ishtml) && (!isfile) && (html_inserted == 0))
					{
						// Does this line have a </body> tag in it... ?
						// - 23/10/2001, AOL 6.0 doesn't have BODY tags
						// 	in it's HTML email, so detect </html> as well
							prebody = strstr(lline,"</body>");
							if (!prebody) prebody = strstr(lline,"</html>");

						if (prebody != NULL)
						{
							html_inserted = 1;
							if (__verbose) fprintf(stdout,"Inserting html-body disclaimer...");

							// prepare to print out the original up to the </body> part
							tmpbody = line +(prebody -lline);
							*tmpbody = '\0';

							// save to file the first part of the line segment
							fprintf(newf,"%s",line);


							if (istext)
							{
								// If the disclaimer is just a text string...
									//
									fprintf(newf,"<PRE>\n%s\n</pre>",disclaimer);
							}
							else
							{
								d = fopen(disclaimer,"r");
								content_type = _text_html;
								if (d)
								{
									AM_insert_disclaimer(d, newf, content_type, content_encoding);
									fclose(d);
								}
								else
								    {
									if (_AM_syslogging > 0) syslog(1,"AM_add_disclaimer: Error: Cannot open disclaimer file for writing (%s) : %s",disclaimer, strerror(errno));
									if (_AM_stderrlogging > 0) fprintf(stderr,"AM_add_disclaimer: Error: Cannot open disclaimer file for reading");
								}
							}

							fprintf(newf,"<%s",(tmpbody+1));

							line[0] = '\0';
							safe[0] = '\0';
							break;
						} // If prebody != NULL
					} // If this is HTML and it's not a file

					fprintf(newf,"%s",line);

				} /* while */
			} /* if HTML too */
			continue;
		} /* if-ELSE (boundary) */

		fprintf(newf,"%s",line);

		/* if our disclaimer was inserted, then we will just runn out the last of
		* the MIMEpack lines, so we dont need to process them with strstr()'s as
		* we've already performed the required task */
		if (disclaimer_inserted)
		{
			while (fgets(line,1024,mp)) fprintf(newf,"%s",line);
		}

	} /* WHILE more lines to read in from original */

	/* if we didn't get the disclaimer inserted, then we've encountered a plain-text email
	* which we just need to insert the disclaimer at the end of it */


	if (!disclaimer_inserted)
	{
		if (__verbose) fprintf(stderr,"No disclaimer added, despite going through entire file...\n");

		if (istext)
		{
			fprintf(newf,"\n%s\n\n",disclaimer);
		}
		else
		    {
			d = fopen(disclaimer,"r");
			if (!d)
			{
				fprintf(stderr,"Error Cannot open disclaimer file %s,%s", disclaimer, strerror(errno));
			}
			else
			    {
				AM_insert_disclaimer(d, newf, content_type, content_encoding);
				fclose(d);
			}
		}
	}

	if (__verbose) fprintf(stdout,"done.\n");

	/* Close our files */
	fclose(mp);
	fclose(newf);

	/* rename the files */
	//rename(mpackname, mpackold);
	remove(mpackname);
	rename(mpacktmp, mpackname);


	return result;
}






/*------------------------------------------------------------------------
Procedure:     AM_filename_fix ID:1
Purpose:       Converts a MIME header
Input:         line: string with the filename
removed_prefix: filename prefix to use
removed_count: the number of attachments we've processed / ID
Output:
Errors:
------------------------------------------------------------------------*/
int AM_filename_fix( char *line, char *removed_prefix, int removed_count )
{
	char *pos;
	char lline[1024];

	/* process the current line, as it contains the filename, and replace the filename
	* with a -removed- one */
	pos = strrchr(line,'='); /* locate where the filename starts */
	pos++; /* move one char along */
	*pos = '\0'; /* terminate the string */

	sprintf(lline,"%s\"%s%d\"\n",line,removed_prefix,removed_count); /* create a new string, in lline (scratch pad)*/
	strcpy(line,lline);

	return 0;
}






/*------------------------------------------------------------------------
Procedure:     AM_nullify_attachment ID:1
Purpose:       Removes any attachment within the mailpack which matches
(minimally) with the attachmentname parameter
Input:         mpackname - The mailpack name
attachmentname - The name of attachments which is to be removed
Output:        none
Errors:        If attachment wasn't found, or could not be removed, a non-zero return value.
------------------------------------------------------------------------*/
int AM_nullify_attachment( char *mpackname, char *attachmentname, char *removed_prefix )
{
	FILE *mp, *newf;
	char line[1024]="";
	char safe[1024]="";
	char lline[1024]="";
	char mpacktmp[1024]="";
	char mpackold[1024]="";
	char boundary[1024]="";

	int content_type = _none;
	int content_encoding = _none;

	int result = 0;
	int boundary_found = 0;
	int filename_found = 0;
	int removed_count = 0;
	char *pos, *newpos;

	if (__verbose) fprintf(stdout,"Removing attachments \"%s\"...",attachmentname);

	/* create our temp filename */
	sprintf(mpacktmp,"%s.tmp",mpackname);
	sprintf(mpackold,"%s.old",mpackname);

	/* Initialise our files */
	newf = fopen(mpacktmp,"w");

	if (!newf)
	{
		if (_AM_syslogging > 0) syslog(1,"AM_nullifyattachment: Error: Cannot open temporary file for writing (%s) : %s",mpacktmp, strerror(errno));
		if (_AM_stderrlogging > 0) fprintf(stderr,"AM_nullifyattachment: Error: Cannot open temporary file for writing");
		return -1;
	}

	mp = fopen(mpackname,"r");
	if (!mp)
	{
		if (_AM_syslogging > 0) syslog(1,"AM_nullifyattachment: Error: Cannot open temporary file for reading (%s) : %s",mpackname, strerror(errno));
		if (_AM_stderrlogging > 0) fprintf(stderr,"AM_nullifyattachment: Error: Cannot open temporary file for reading");
		return -1;
	}

	/* Initialise our variables */
	content_type = _none;
	content_encoding = _none;

	AM_strlower(attachmentname);

	/* Read every line from the original mailpack */
	while ( fgets(line,1024,mp) )
	{

		/* copy and convert into lower-case the input line */
		strcpy(lline,line);
		strcpy(safe,line);
		AM_strlower(lline);

		/* check and see if we've found the filename... */

		if ( (pos=strstr(lline, "name="))!=NULL)
		{
			if (  (__nullify_all) || (pos=strstr(lline,attachmentname))  )
			{
				filename_found = 1;
				if (__verbose) fprintf(stdout,"\nlocated attachment...");
			}
		}

		/* check and see if this is a content-type line */
		if ((pos=strstr(lline, "content-type:")))
		{
			pos += strlen("content-type:");	/* move our pointer along */

			if ((pos=strstr(pos,"text/")))	/* do we have a text type */
			{
				pos += strlen("text/");	/* move pointer along */

				if (strncmp(pos, "plain",5)==0) content_type = _text_plain;
				else
				    if (strncmp(pos, "html",4)==0) content_type = _text_html;
				else
				    if (strncmp(pos, "print",5)==0) content_type = _text_print;
				else content_type = _text_unknown;

			}
			else {
				content_type = _unknown;
			}
		}

		/* Our transfer encoding system indicates if we have a attachment type which is
		* binary-encoded stuff or plain text */
		if ((pos=strstr(lline, "content-transfer-encoding:")))
		{
			pos += strlen("content-transfer-encoding:");
			if (strstr(pos,"base64")) content_encoding = _base64;
			else if (strstr(pos,"quote")) content_encoding = _print;
			else if (strstr(pos,"7bit")) content_encoding = _7bit;
		}


		if (boundary_found < 1)
		{ /* if we have not found our boundary yet */
			pos = strstr(lline, "boundary=");	/* check for our boundary */
			if (pos) { /* if we found one! */

				newpos = (pos -lline) +line +strlen("boundary=");	/* move the pos pointer to just after the = */
				pos = newpos;
				if (*pos == '\"') pos++;	/* if our 1st char is a quote, move one more */
				newpos = strtok(pos,"\n\r\"");	/* tokenise out the boundary */
				strcpy(boundary, newpos);	/* copy it to our static string array */
				boundary_found++;	/* increment the boundary_found flag */
				if (__verbose) fprintf(stdout,"found boundary...");
			} /* if POS */
		} /* if boundary_found < 1 */

		else
		    { /* if boundary_found >= 1 */

			/* if we've found the filename, then we must process until we find the next boundary */
			if (filename_found == 1)
			{
				AM_filename_fix(line, removed_prefix, removed_count);
				fprintf(newf,"%s",line); /* output the new filename */
				if (__verbose) fprintf(stdout,"renamed attachment...");

				/* first, locate our \n\r blank line */

				while ( (fgets(line, 1024,mp)) && (strlen(line) >3) ){
					strcpy(lline,line);
					AM_strlower(lline);

					if (strstr(lline,"name=")){
						AM_filename_fix(line, removed_prefix, removed_count);
					}

					fprintf(newf,"%s",line);
				}


				/* Now, until the boundary is found again, absorb all lines */
				while ( (fgets(line, 1024,mp)) && (!strstr(line, boundary)) );

				/* leave an empty file */
				fprintf(newf,"\n\n");

				/* and, given that we will exit on the boundary line, we can just edit from this loop as we write it later */
				strcpy(safe, line);

				/* we are -done- */
				filename_found = 0;
				removed_count++;
				if (__verbose) fprintf(stdout,"nullified.");
			} /* if filename_found == 1 */

		} /* if-ELSE */

		fprintf(newf,"%s",safe); /* save the original line to the output file */
		filename_found = 0;

	} /* WHILE more lines to read in from original */

	if (__verbose) fprintf(stdout,"\ndone.\n");
	/* Close our files */
	fclose(mp);
	fclose(newf);

	/* Rename files */
	rename(mpackname, mpackold);
	rename(mpacktmp, mpackname);

	return result;
}


/*---------------------------------END-*/



/*----------------------------------------------------------------------
* MIME_insert_X_header
*
* Inserts a specified X-header at the end of other X-headers.
*/
int AM_insert_Xheader( char *fname, char *xheader)
{

	/* Tempfile tmpfile() fix contributed by David DeMaagd - 29/01/2001 */
	char line[1024], lline[1024];
	char tpn[1024];
	int xon = 0;
	int returnvalue = 0;
	int _AM_syslogging = 0;
	int _AM_stderrlogging = 1;
	FILE *fo;
	FILE *f;

	/* Get temp name */
	sprintf(tpn,"%s.",fname);

	fo = fopen(tpn,"w");
	if (!fo)
	{
		if (_AM_syslogging > 0) syslog(1,"Error: Cannot open temporary file for writing (%s) : %s",tpn, strerror(errno));
		if (_AM_stderrlogging > 0) fprintf(stderr,"Error: Cannot open temporary file for writing");
		returnvalue = 1;
	}
	f = fopen(fname,"r");
	if (!f)
	{
		if (_AM_syslogging > 0) syslog(1,"Error: Cannot open Mailpack file for reading (%s) : %s",fname, strerror(errno));
		if (_AM_stderrlogging > 0) fprintf(stderr,"Error: Cannot open Mailpack file for reading, %s",fname);
		returnvalue = 2;
	}

	if ( returnvalue == 0 )
	{
		while (fgets(line,1023,f))
		{
			strncpy(lline,line,1023);
			AM_strlower(lline);

			if ((xon==0)&&(strstr(lline,"content-type:")))
			{
				fprintf(fo,"%s\n",xheader);
				xon=1;
			}

			fprintf(fo,"%s",line);
		}

		/* close files */
		fclose(fo);
		fclose(f);

		/* remove the old file */
		if (remove(fname) == -1) {
			if (_AM_stderrlogging > 0) fprintf(stderr,"Error: xInsertHeader - REMOVE, %s",strerror(errno));
		}
		else {
			/* and rename */
			if (rename(tpn,fname) == -1) {
				if (_AM_stderrlogging > 0) fprintf(stderr,"Error: xInsertHeader - RENAME, %s",strerror(errno));
			} /*if rename*/
		} /* else */

	} /* if */

	return returnvalue;
}





//-----------------------------------END----


